import io
import json
import threading
import time
import requests
import logging

DEFAULT_REQUESTS_PER_MINUTE_LIMIT = 900
logger = logging.getLogger(__name__)


class IrysException(Exception):
    pass


class IrysNetworkException(IrysException):
    pass


class HTTPClient:
    def __init__(
        self,
        api_url,
        timeout=None,
        retries=10,
        outgoing_connections=256,
        requests_per_period=DEFAULT_REQUESTS_PER_MINUTE_LIMIT,
        period_sec=60,
        extra_headers={},
        cert_fingerprint=None,
    ):
        self.api_url = api_url
        self.session = requests.Session()
        self.max_outgoing_connections = outgoing_connections
        self.outgoing_connection_semaphore = threading.BoundedSemaphore(
            outgoing_connections
        )
        self.rate_limit_lock = threading.Lock()
        self.requests_per_period = requests_per_period
        self.ratelimited_requests = 0
        self.period_sec = period_sec
        # self.incoming_port = incoming_port
        self.extra_headers = extra_headers
        self.req_history = []
        max_retries = requests.adapters.Retry(
            total=retries, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504]
        )  # from so
        https_adapter = self._FingerprintAdapter(
            fingerprint=cert_fingerprint,
            pool_connections=outgoing_connections,
            pool_maxsize=outgoing_connections,
            max_retries=max_retries,
            pool_block=True,
        )

        http_adapter = requests.adapters.HTTPAdapter(
            pool_connections=outgoing_connections,
            pool_maxsize=outgoing_connections,
            max_retries=max_retries,
            pool_block=True,
        )
        self.session.mount("http://", http_adapter)
        self.session.mount("https://", https_adapter)

        self.timeout = timeout

    def __del__(self):
        self.session.close()

    class _FingerprintAdapter(requests.adapters.HTTPAdapter):
        def __init__(self, fingerprint=None, **kwparams):
            self.fingerprint = fingerprint
            super().__init__(**kwparams)

        def init_poolmanager(self, connections, maxsize, block=False):
            self.poolmanager = requests.packages.urllib3.poolmanager.PoolManager(
                num_pools=connections,
                maxsize=maxsize,
                block=block,
                assert_fingerprint=self.fingerprint,
            )

    def ratelimited(self):
        if self.requests_per_period is None:
            return False
        with self.rate_limit_lock:
            now = time.time()
            queued_requests = 0
            for idx, then in enumerate(self.req_history):
                if then + self.period_sec >= now:
                    queued_requests_idx = idx
                    queued_requests = len(
                        self.req_history) - queued_requests_idx
                    break
            return (
                queued_requests + self.ratelimited_requests >= self.requests_per_period
            )

    def ratelimit_suggested(self):
        if self.requests_per_period is None:
            return False
        if len(self.req_history) == 0:
            return False
        return (
            time.time() - self.req_history[-1]
            < self.requests_per_period / self.period_sec
        ) or self.ratelimited

    def _ratelimit_prologue(self):
        if self.requests_per_period is None:
            return
        with self.rate_limit_lock:
            now = time.time()
            queued_requests = 0
            # if len(self.req_history) >= 3600:
            #    import pdb; pdb.set_trace()
            for idx, then in enumerate(self.req_history):
                if then + self.period_sec >= now:
                    queued_requests_idx = idx
                    queued_requests = len(
                        self.req_history) - queued_requests_idx
                    break
            if queued_requests + self.ratelimited_requests < self.requests_per_period:
                # if len(self.req_history) >= self.requests_per_period:
                #    import pdb; pdb.set_trace()
                self.req_history.append(now)
                return
        # print(f'{self.api_url}: too many requests in prologue')
        self.on_too_many_requests()
        with self.rate_limit_lock:
            now = time.time()
            if len(self.req_history) >= self.requests_per_period:
                duration = (
                    self.req_history[-self.requests_per_period + 1]
                    + self.period_sec
                    - now
                )
                if duration > 0:
                    if duration > 0.5:
                        # quick workaround to let this display later during lock contention
                        time.sleep(0.5)
                        duration -= 0.5
                    logger.info(
                        f"Sleeping for {int(duration*100)/100}s to respect ratelimit of {
                            self.requests_per_period}req/{self.period_sec}s ..."
                    )
                    time.sleep(duration)
                    # import pdb; pdb.set_trace()
                    logger.info(
                        f"Done sleeping for {int(duration*100)/100}s to respect ratelimit of {
                            self.requests_per_period}req/{self.period_sec}s ."
                    )
        return self._ratelimit_prologue()

    def _ratelimit_epilogue(self, success=True):
        if self.requests_per_period is None:
            return
        if success:
            with self.rate_limit_lock:
                self.ratelimited_requests = 0
                now = time.time()
                for req_idx, req_time in enumerate(self.req_history):
                    if req_time + self.period_sec > now:
                        self.req_history = self.req_history[req_idx:]
                        break
        else:
            with self.rate_limit_lock:
                self.ratelimited_requests += 1
                now = time.time()
                # import pdb; pdb.set_trace()
                if len(self.req_history):
                    self.period_sec = max(
                        self.period_sec, now - self.req_history[0])
                if (
                    len(self.req_history) - self.ratelimited_requests
                    <= self.requests_per_period
                ):
                    self.requests_per_period = max(
                        1, len(self.req_history) - self.ratelimited_requests
                    )
                logger.info(
                    f"Rate limit hit. Dropped rate to {
                        self.requests_per_period}/{self.period_sec}s."
                )
            self.on_too_many_requests()

    # _get and _post should just call a _request function to share code
    def _request(self, *params, **request_kwparams):
        if len(params) and params[-1][0] == "?":
            url = self.api_url + "/" + "/".join(params[:-1]) + params[1]
        else:
            url = self.api_url + "/" + "/".join(params)

        headers = {**self.extra_headers, **request_kwparams.get("headers", {})}
        request_kwparams["headers"] = headers

        while True:
            self._ratelimit_prologue()
            response = None
            try:
                if not self.outgoing_connection_semaphore.acquire(blocking=False):
                    self.on_too_many_connections()
                    logger.info(
                        "Waiting for connection count limit semaphore to drain..."
                    )
                    self.outgoing_connection_semaphore.acquire()
                try:
                    response = self.session.request(
                        **{"url": url, "timeout": self.timeout, **request_kwparams}
                    )
                finally:
                    self.outgoing_connection_semaphore.release()

                if response.status_code == 400:
                    try:
                        msg = response.json()["error"]
                    except:
                        msg = response.text
                    raise IrysException(msg)

                response.raise_for_status()
                if int(response.headers.get("content-length", 1)) == 0:
                    raise IrysException(f"Empty response from {url}")
                self._ratelimit_epilogue(True)
                return response
            except requests.exceptions.RequestException as exc:
                text = "" if response is None else response.text
                status_code = 0 if response is None else response.status_code
                if status_code == 429:
                    # too many requests
                    self._ratelimit_epilogue(False)
                    self.on_too_many_requests()
                    continue
                if (
                    type(exc) is requests.ConnectionError
                    and len(exc.args) > 0
                    and type(exc.args[0]) is requests.urllib3.exceptions.ClosedPoolError
                ):
                    # strange ClosedPoolError from urllib3 race condition? https://github.com/urllib3/urllib3/issues/951
                    self._ratelimit_epilogue(False)  # to reduce busylooping
                    continue
                if type(exc) is requests.ReadTimeout:
                    if status_code == 0:
                        status_code = 598
                    logger.info("{}\n{}\n\n{}".format(
                        exc, text, request_kwparams))
                else:
                    pass
                if status_code == 520:
                    # cloudfront broke
                    self._ratelimit_epilogue(True)
                    continue
                elif status_code == 502:
                    # cloudflare broke
                    self._ratelimit_epilogue(True)
                    continue
                self.on_network_exception(text, status_code, exc, response)
                raise IrysNetworkException(
                    text or repr(type(exc)), status_code, exc, response
                )
            except:
                self._ratelimit_epilogue(True)
                raise

    def _get(self, *params, **request_kwparams):
        return self._request(*params, **{"method": "GET", **request_kwparams})

    def _get_json(self, *params, **request_kwparams):
        response = self._get(*params, **request_kwparams)
        try:
            return response.json()
        except:
            raise IrysException(response.text)

    def _post(self, data, *params, headers={}, **request_kwparams):
        headers = {**headers}

        if type(data) is dict:
            headers.setdefault("Content-Type", "application/json")
            data_key = "json"
        else:
            if isinstance(data, (bytes, bytearray)):
                headers.setdefault("Content-Type", "application/octet-stream")
            else:
                headers.setdefault("Content-Type", "text/plain")
            data_key = "data"

        return self._request(
            *params,
            **{
                "method": "POST",
                "headers": headers,
                **{data_key: data},
                **request_kwparams,
            },
        )

    def _post_json(self, data, *params, **request_kwparams):
        response = self._post(data, *params, **request_kwparams)
        try:
            return response.json()
        except json.decoder.JSONDecodeError:
            raise IrysException(response.text)

    def on_network_exception(self, text, code, exception, response):
        text = f"{self.api_url} - {text}"
        raise IrysNetworkException(text, code, exception, response)

    def on_too_many_connections(self):
        pass

    def on_too_many_requests(self):
        return self.on_too_many_connections()



from ritual_irys.http import HTTPClient
from ritual_irys.utils import DEFAULT_GATEWAY
DEFAULT_REQUESTS_PER_MINUTE_LIMIT = 900


class Node(HTTPClient):

    def __init__(
        self,
        api_url=DEFAULT_GATEWAY,
        timeout=None,
        retries=5,
        outgoing_connections=DEFAULT_REQUESTS_PER_MINUTE_LIMIT,
        requests_per_period=DEFAULT_REQUESTS_PER_MINUTE_LIMIT,
        period_sec=60,
        incoming_port=None,
    ):
        super().__init__(
            api_url,
            timeout,
            retries,
            outgoing_connections,
            requests_per_period,
            period_sec,
            extra_headers={"X-P2p-Port": str(incoming_port)}
            if incoming_port is not None
            else {},
        )

    def tx_field(self, hash, field):
        """
        Return a given field of the transaction specified by the transaction ID (hash).

        {field} := {
            'id' | 'last_tx' | 'owner' | 'tags' | 'target' | 'quantity' |
            'data_root' | 'data_size' | 'data' | 'reward' | 'signature'
        }
        """
        match field:
            case "data":
                response = self._get("tx", hash, "data")
                return response.content
            # case "data_size":
            #     response = self._get_json("tx", hash)
            #     return response["data_size"]
            case _:
                response = self._get_json("tx", hash)
                # for tag in response:
                #     for key in tag:
                #         tag[key] = b64dec(tag[key].encode())
                return response[field]

    def tx_meta(self, hash):
        self._get_json("tx", hash)

    def tx_id(self, hash):
        """Return transaction id."""
        return self.tx_field(hash, "id")

    def tx_owner(self, hash):
        """Return transaction owner."""
        return self.tx_field(hash, "owner")

    def tx_tags(self, hash):
        """Return transaction tags."""
        return self.tx_field(hash, "tags")

    def tx_target(self, hash):
        """Return transaction target."""
        return self.tx_field(hash, "target")

    # def tx_quantity(self, hash):
    #     """Return transaction quantity."""
    #     return self.tx_field(hash, "quantity")

    # def tx_data_root(self, hash):
    #     """Return transaction data root."""
    #     return self.tx_field(hash, "data_root")

    def tx_data_size(self, hash):
        """Return transaction data size."""
        return int(self.tx_field(hash, "data_size"))

    def tx_data(self, hash):
        """
        Return transaction data.

        The endpoint serves data regardless of how it was uploaded.
        """
        return self.tx_field(hash, "data")

    def tx_signature(self, hash):
        """Return transaction signature."""
        return self.tx_field(hash, "signature")

    # def tx_reward(self, hash):
    #     """Return transaction reward."""
    #     return self.tx_field(hash, "reward")

    # def height(self):
    #     """Return the current block hieght."""
    #     response = self._get("height")
    #     return int(response.text)

    def data(self, txid, ext="", range=None, timeout=60):
        """
        Get the decoded data from a transaction.

        This is roughly just an alias for tx_data_html.

        The transaction is pending: Pending
        The provided transaction ID is not valid or the field name is not valid: Invalid hash.
        A transaction with the given ID could not be found: Not Found.
        """
        if range is not None:
            headers = {"Range": f"bytes={range[0]}-{range[1]}"}
        else:
            headers = {}
        response = self._get(txid + ext, headers=headers, timeout=timeout)

        return response.content

    def graphql(self, query):
        response = self._post_json(
            {"operationName": None, "query": query, "variables": {}}, "graphql"
        )
        return response



import concurrent.futures
import logging
import queue
from queue import Queue
from typing import Any, Callable, Generic, List, Optional, Type, TypeVar, cast

log = logging.getLogger(__name__)

WorkerContext = TypeVar("WorkerContext")
TaskContext = TypeVar("TaskContext")


class Worker(Generic[WorkerContext, TaskContext]):
    """
    A worker that processes tasks from a queue.

    Args:
        context (WorkerContext): The context for the worker.
        work (Callable[[WorkerContext, TaskContext], None]): A function that processes
            a task.
        evict_hook (Callable[[WorkerContext], None]): A function that is called when
            the worker is evicted.

    """

    def __init__(
        self,
        context: WorkerContext,
        work: Callable[[WorkerContext, TaskContext], None],
        evict_hook: Callable[[WorkerContext], None] = lambda x: None,
    ):
        """
        Initialize the Worker with a context, a work function, and an evict hook.

        Args:
            context (WorkerContext): The context for the worker.
            work (Callable[[WorkerContext, TaskContext], None]): A function that
                processes a task.
            evict_hook (Callable[[WorkerContext], None]): A function that is called when
                the worker is evicted.
        """
        self.context = context
        self.work = work
        self.evict_hook = evict_hook


class QueueProcessor(Generic[WorkerContext, TaskContext]):
    """
    A queue with a list of workers that process tasks from the queue.
    If a worker fails, the task is re-added to the queue, and the worker is evicted.
    Another worker will pick up the task. This is useful for processing tasks in parallel
    where some tasks may fail due to network issues or other reasons.
    """

    def __init__(
        self, evict_exceptions: List[Type[Exception]] | None = None, logger: Any = log
    ):
        """
        Initialize the QueueProcessor with a queue and a list of workers.

        Args:
            evict_exceptions (List[Type[Exception]]): A list of exceptions upon which
            workers from the worker list get evicted.
            logger (Any): The logger to use for logging.
        """
        if evict_exceptions is None:
            evict_exceptions = []

        self.queue: Queue[TaskContext] = Queue()
        self.workers: List[Worker[WorkerContext, TaskContext]] = []
        self.evict_exceptions = evict_exceptions
        self.log = logger

    def add_task(self, task: TaskContext) -> None:
        """
        Add a task to the queue.

        Args:
            task (TaskContext): The task to add.

        Returns:
            None
        """
        self.queue.put(task)

    def add_worker(self, worker: Worker[WorkerContext, TaskContext]) -> None:
        """
        Add a worker to the list of workers.

        Args:
            worker (Worker[WorkerContext, TaskContext]): The worker to add.

        Returns:
            None
        """
        self.workers.append(worker)

    def _start_worker_loop(self, worker: Worker[WorkerContext, TaskContext]) -> None:
        """
        Start a worker loop and process the queue until it is empty.

        Args:
            worker (Worker[WorkerContext, TaskContext]): The worker to use.

        Returns:
            None
        """
        while not self.queue.empty():
            task: Optional[TaskContext] = None
            try:
                task = self.queue.get(block=False)
                worker.work(worker.context, task)
                self.queue.task_done()
            except queue.Empty:
                break
            except Exception as e:
                self.queue.put(cast(TaskContext, task))
                self.queue.task_done()
                raise e

    def _exhaust_queue_with_worker(
        self, worker: Worker[WorkerContext, TaskContext]
    ) -> None:
        """
        Start a worker loop and process the queue until it is empty.

        Args:
            worker (Worker[WorkerContext, TaskContext]): The worker to use.

        Returns:
            None
        """
        while self.queue.unfinished_tasks > 0:
            try:
                self._start_worker_loop(worker)
            except Exception as e:
                self.log.error(
                    f"Worker: {worker} failed, no longer using it: {e}"
                    f" - unfinished tasks: {self.queue.unfinished_tasks}"
                )
                worker.evict_hook(worker.context)
                self.log.debug(f"Error: {e}")
                break

    def process(self) -> None:
        """
        Start a thread pool with the workers and process the queue.
        """
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=len(self.workers)
        ) as executor:
            futures = [
                executor.submit(self._exhaust_queue_with_worker, worker)
                for worker in self.workers
            ]
            for future in futures:
                future.result()



"""
Script containing a command line interface to upload and download repositories from/to
irys. This file is a thin wrapper around RepoManager & FileManager classes to
facilitate usage of the repo upload and download functions from the command line.

This CLI is commonly used to upload and download model files from/to irys, similar
to how models are handled in huggingface hub.
"""

import json
from pathlib import Path
from typing import Any, Callable, Optional

import click
from ritual_irys.file_manager import FileManager
from ritual_irys.repo_manager import RepoManager
from ritual_irys.utils import DEFAULT_BUNDLER_NODE, DEFAULT_GATEWAY


@click.group()
def cli() -> None:
    pass


GenericCallable = Callable[..., Any]


# Define reusable options as functions
def repo_name_option(f: GenericCallable) -> GenericCallable:
    return click.option(
        "--repo-name",
        type=str,
        required=True,
        help="Repo name to upload. Once you upload this repo, it can be downloaded "
        "using the repo id: `owner/repo_name` where `owner` is the wallet address"
        "that uploaded the repo.",
    )(f)


def repo_id_option(f: GenericCallable) -> GenericCallable:
    return click.option(
        "--repo-id",
        type=str,
        required=True,
        help="Repo ID to upload or download. This is a unique identifier for the repo,"
        " and is in the format: `owner/repo_name` where `owner` is the wallet "
        "address that uploaded the repo, and `repo_name` is the name of the "
        "repo.",
    )(f)


def wallet_option(f: GenericCallable) -> GenericCallable:
    return click.option(
        "--wallet",
        type=str,
        required=False,
        default="wallet.txt",
        help="Path to wallet file, default is `wallet.txt`",
    )(f)


def bundler_option(f: GenericCallable) -> GenericCallable:
    return click.option(
        "--bundler",
        type=str,
        required=False,
        default=DEFAULT_BUNDLER_NODE,
        help=f"Irys bundler node to use, default is uploader.irys.xyz "
        f"{DEFAULT_BUNDLER_NODE}",
    )(f)


def gateway_option(f: GenericCallable) -> GenericCallable:
    return click.option(
        "--gateway",
        type=str,
        required=False,
        default=DEFAULT_GATEWAY,
        help=f"Irys gateway to use, default is "
        f"{DEFAULT_GATEWAY}",
    )(f)


def max_upload_size_option(f: GenericCallable) -> GenericCallable:
    return click.option(
        "--max-upload-size",
        type=int,
        required=False,
        default=None,  # Assuming a default value is set
        help="If a file is too large to be uploaded in a single transaction, it will be "
        "split into chunks of this size. Default is 5MB",
    )(f)


def show_progress_option(f: GenericCallable) -> GenericCallable:
    return click.option(
        "--show-progress",
        type=bool,
        required=False,
        default=True,
        help="If set, it will show progress bar while uploading files",
    )(f)


def token_option(f: GenericCallable) -> GenericCallable:
    return click.option(
        "--token",
        type=str,
        required=False,
        default="ethereum",
        help="Token to use, default is `ethereum`",
    )(f)


@click.option(
    "--repo-dir",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Enter repo directory",
)
@click.option(
    "--version-file",
    required=False,
    default=None,
    type=str,
    help="enter version mapping json of repo files to upload",
)
@repo_name_option
@wallet_option
@show_progress_option
@bundler_option
@token_option
@cli.command(
    name="upload-repo",
)
def upload_repo(
    repo_name: str,
    repo_dir: str,
    version_file: Optional[str],
    wallet: str,
    show_progress: bool,
    bundler: str,
    token: str
) -> None:
    """
    Uploads a repo to Irys using the specified repo name & repo directory.

    Parameters:
        repo_name: Name of the repo to upload.
        repo_dir: Path to the repo directory.
        version_file (optional): Path to the version mapping file. This is a json file
            that maps repo filenames to their corresponding versions.
        wallet (optional): Private key or path to the wallet file. Default is `wallet.txt`.
        show_progress: If set, it will show progress bar while uploading files.
        gateways (optional): Commas separated list of Irys gateways.


    Examples:

    To upload a repo with ID <repo-id> from the directory <repo-dir>:

    ritual-irys upload-repo --repo-name <repo-name> --repo-dir <repo-dir>

    To upload a repo with ID <repo-id> from the directory <repo-dir> and version
    mapping file <version-file>:

    ritual-irys upload-repo --repo-name <repo-name> --repo-dir <repo-dir> \
        --version-file <version-file>

    To upload a repo with ID <repo-id> from the directory <repo-dir> and
    wallet <wallet>:

    ritual-irys upload-repo --repo-name <repo-name> --repo-dir <repo-dir> \
        --wallet <wallet>

    """
    r = RepoManager(
        bundler_url=bundler,
        wallet=wallet,
        token=token,
        show_progress_bar=show_progress,
    ).upload_repo(name=repo_name, path=repo_dir, version_mapping_file=version_file)
    click.echo(
        f"uploaded repo: {r}"
        f"\n\tyou can download it using the repo id: "
        f"`{r.repo_id.owner}/{r.repo_id.name}`"
    )


@click.option(
    "--base-path",
    type=str,
    default=".",
    required=False,
    help="enter base path to save repo files, defaults to the current directory.",
)
@click.option(
    "--force-download",
    is_flag=True,
    default=False,
    required=False,
    help="If set, it will override the existing repo files if they exist",
)
@repo_id_option
@show_progress_option
@gateway_option
@cli.command(
    name="download-repo",
)
def download_repo(
    repo_id: str,
    base_path: str = ".",
    force_download: bool = False,
    show_progress: bool = True,
    gateway: str = DEFAULT_GATEWAY,
) -> None:
    """
    Downloads a repo from Irys using the specified repo ID, and
    API URL. Optionally, you can specify multiple owners and a
    base path where the repo files will be saved.
    Use the --force-download flag to override existing files.

    Parameters:
        repo_id: The ID of the repository to download.
        base_path: The base path to save the repo files. Defaults to the current
            directory.
        force_download: If set, it will override the existing repo files if they exist.
        show_progress: If set, it will show progress bar while downloading files.
        gateways: Commas separated list of Irys gateways.

    Examples:

    To download a repo with ID <repo-id> and owner <owner-address>:

    ritual-irys download-repo --repo-id <repo-id> --owner <owner-address>

    To download a repo with ID <repo-id> and owner <owner-address> and save the repo
    files to <path-to-save-repo>:

    ritual-irys download-repo --repo-id <repo-id> --base-path <path-to-save-repo>
    """

    files = RepoManager(
        gateway_url=gateway, show_progress_bar=show_progress
    ).download_repo(
        repo_id,
        base_path,
        force_download,
    )

    click.echo(f"downloaded files: {files}")


@click.option(
    "--file-path",
    type=str,
    help="Path to the file to download",
)
@click.option(
    "--tx-id",
    type=str,
    help="Transaction id of the file to download",
)
@show_progress_option
@gateway_option
@cli.command(
    name="download-file",
)
def download_file(
    file_path: str,
    tx_id: str,
    gateway: str = DEFAULT_GATEWAY,
    show_progress: bool = True,
) -> None:
    """
    Downloads a file from Irys using the transaction ID.

    Parameters:
        file_path: Path to the file to download.
        tx_id: Transaction ID of the file to download.
        bundler: URL of the bundler node to use.
        show_progress: If set to false, it will not show the progress bar while

    Examples:

    To download a file with transaction id <tx-id> to the path <file-path>:

    ritual-irys download-file --file-path <file-path> --tx-id <tx-id>
    """
    fm = FileManager(gateway_url=gateway, show_progress_bar=show_progress)
    fm.download(file_path, tx_id)


@click.option(
    "--file-path",
    type=str,
    help="Path to the file to upload",
    required=True,
)
@click.option(
    "--tags",
    type=str,
    required=False,
    default="{}",
    help="Dictionary of tags to attach to the file. Must be a JSON string.",
)
@max_upload_size_option
@show_progress_option
@bundler_option
@token_option
@wallet_option
@cli.command(
    name="upload-file",
)
def upload_file(
    file_path: str,
    max_upload_size: int,
    token,
    wallet,
    show_progress: bool = True,
    bundler: str = DEFAULT_BUNDLER_NODE,
    tags: str = "{}",
) -> None:
    """
    Uploads a file to Irys.

    Parameters:
        file_path: Path to the file to upload.
        max_upload_size: For large files, the file will be split into chunks of this
        size.
        bundler: URL of the bundler node to use.
        wallet: Private key or path to the wallet file. Default is `wallet.txt`
        tags: Dictionary of tags to attach to the file. Must be a JSON string.

    Examples:

    To upload a file with path <file-path>:

    ritual-irys upload-file --file-path <file-path>
    """
    fm = FileManager(
        bundler_url=bundler,
        wallet=wallet,
        token=token,
        max_upload_size=max_upload_size,
        show_progress_bar=show_progress,
    )
    tx = fm.upload(Path(file_path), tags_dict=json.loads(tags))
    click.echo(f"uploaded file: {tx.id}")



import json

# from .utils import create_tag

MANIFEST_CONTENT_TYPE = 'application/x.irys-manifest+json'


class Manifest:
    def __init__(self, path_txids, index=None, version='0.1.0'):
        self.path_txids = path_txids
        self.index = index
        self.version = version

    @classmethod
    def fromjson(cls, json):
        return cls(
            path_txids={key: value['id']
                        for key, value in json['paths'].items()},
            index=json['index']['path'] if 'index' in json else None,
            version=json['version']
        )

    def tojson(self):
        return dict(
            manifest='irys/paths',
            version=self.version,
            **(
                dict(
                    index=dict(
                        path=self.index
                    )
                ) if self.index is not None
                else {}
            ),
            paths={
                path: txid if type(txid) is dict else dict(id=txid)
                for path, txid in self.path_txids.items()
            }
        )

    def tobytes(self):
        return json.dumps(self.tojson()).encode()

    @staticmethod
    def totags():
        return [('Content-Type', MANIFEST_CONTENT_TYPE), ("Type", "manifest")]



from __future__ import annotations

from typing import Dict, NewType

from pydantic import BaseModel

IrysAddress = NewType(
    "IrysAddress",
    str,
)
"""Type alias for Irys addresses, represented as strings."""

IrysTransactionId = NewType("IrysTransactionId", str)
"""Type alias for Irys transaction IDs, represented as strings."""

Tags = Dict[str, str]
"""Type alias for a dictionary of tags, where both keys and values are strings."""


class IrysRepoId(BaseModel):
    """
    A class representing a repository identifier on Irys.

    Attributes:
        owner (IrysAddress): The owner of the repository.
        name (str): The name of the repository.
    """

    owner: IrysAddress
    name: str

    @classmethod
    def from_str(cls, _id: str) -> IrysRepoId:
        """
        Create a IrysRepoId instance from a string.

        Args:
            _id (str): A string in the format 'owner/name'.

        Returns:
            IrysRepoId: An instance of IrysRepoId with the owner and name extracted from the
            input string.
        """
        owner, name = _id.split("/")
        return cls(owner=IrysAddress(owner), name=name)


class LargeFileManifest(BaseModel):
    """
    A class to represent a manifest file for a large file uploaded in sections.

    Attributes:
        files (Dict[str, str]): A dictionary of file names and their transaction IDs.
        size (int): The total size of the file.
    """

    size: int
    files: Dict[str, str]



"""
Utility functions for the Irys integration.
"""

from irys_sdk.builder import Builder
from irys_sdk.client import Uploader
from irys_sdk.bundle.tags import Tags
import hashlib
import logging
from typing import Any, Dict


DEFAULT_BUNDLER_NODE = "https://uploader.irys.xyz"
DEFAULT_GATEWAY = "https://gateway.irys.xyz"
DEFAULT_TOKEN = "ethereum"


# max single tx size (TODO remove once chunking is implemented)
MAX_TX_BYTES = 50 * 1000 * 1000

log = logging.getLogger(__name__)


def get_tags_dict(tag_dicts: list[dict[str, str]]) -> dict[str, str]:
    """
    Helper function to merge a list of tag dicts into
    a single dictionary.

    Args:
        tag_dicts(list[dict[str, str]): a list of tag dicts with
        keys 'name' and 'value' corresponding to the name and
        value of the tag respectively.

    Returns:
        dict[str, str]: a key value dict mapping tag name to tag value
    """
    tags: dict[str, str] = {item["name"]: item["value"] for item in tag_dicts}
    return tags


def edge_unix_ts(edge: dict[str, Any]) -> float:
    """
    Helper function to extract the unix time stamp from an
    Irys transaction edge. See https://uploader.irys.xyz/graphql for the
    Irys graphql schema.

    Args:
        edge (dict[str, Any]): a transaction edge object

    Returns:
        float: unix timestamp in seconds
    """
    # sort matching manifests by time, get latest
    tag_dicts: list[dict[str, str]] = edge["node"]["tags"]
    return float(get_tags_dict(tag_dicts)["Unix-Time"])


def get_sha256_digest(file_path: str) -> str:
    """Helper function that computes the digest
    of a file in binary mode to handle potentially
    large files.

    Args:
        file_path (str): path to a file

    Returns:
        str: hex string representing the sha256
    """
    h = hashlib.sha256()

    with open(file_path, "rb") as file:
        while True:
            # Reading is buffered, so we can read smaller chunks.
            chunk = file.read(h.block_size)
            if not chunk:
                break
            h.update(chunk)

    return h.hexdigest()


def from_tags_dict(tags: Dict[str, str]) -> Tags:
    return list(tags.items())


def init_irys(
    wallet: str, token: str = "ethereum", bundler_url: str = DEFAULT_BUNDLER_NODE
) -> Uploader:
    wallet2 = try_read_file(wallet)
    return Builder(token).wallet(wallet2).url(bundler_url).build()


def try_read_file(path_or_value: str) -> str:
    try:
        with open(path_or_value, 'r') as f:
            content = f.read()
            return content if content else path_or_value
    except:
        return path_or_value


def ensure_str(value: str | bytes) -> str:
    if isinstance(value, bytes):
        return value.decode('utf-8')
    return value



"""
File Manager: Utility functions to download and upload files to Irys, as well
as check if a file exists on Irys.
"""

from random import randbytes
from irys_sdk.client import Uploader
import json
import logging
import os
import tempfile
import time
from pathlib import Path
from typing import Any, Callable, Dict,  Optional, Tuple, cast

import requests
from irys_sdk.bundle.dataitem import DataItem
from irys_sdk.bundle.create import create_data
from irys_sdk.bundle.sign import sign
from reretry import retry  # type: ignore
from rich.progress import BarColumn, Progress, ProgressColumn, Task
from rich.progress import TaskID as PBarID
from rich.progress import TextColumn, TimeElapsedColumn, TimeRemainingColumn
from ritual_irys.concurrency_utils import QueueProcessor, Worker
from ritual_irys.gateway import Node
from ritual_irys.manifest import MANIFEST_CONTENT_TYPE, Manifest
from ritual_irys.types import LargeFileManifest, Tags
from ritual_irys.utils import (
    DEFAULT_BUNDLER_NODE,
    DEFAULT_GATEWAY,
    DEFAULT_TOKEN,
    MAX_TX_BYTES,
    ensure_str,
    from_tags_dict,
    get_sha256_digest,
    get_tags_dict,
    init_irys,
)
from ritual_irys.utils import log as default_logger
from ritual_irys.version import VERSION


class FileNotReadyException(Exception):
    """Exception raised when a file is not ready for download from Irys."""

    pass


MAXIMUM_UPLOAD_SIZE = MAX_TX_BYTES

log = logging.getLogger(__name__)

Gateway = str
TxID = str
Section = Tuple[Path, TxID]

STANDBY = "💤"


class FileManager:
    """
    A class to manage file operations with Irys, including downloading,
    uploading, and checking file existence.
    """

    def __init__(
        self,
        wallet: str,
        token: str = DEFAULT_TOKEN,
        bundler_url: str = DEFAULT_BUNDLER_NODE,
        gateway_url: str = DEFAULT_GATEWAY,
        logger: Callable[[str], None] = default_logger.info,
        max_upload_size: int = MAXIMUM_UPLOAD_SIZE,
        show_progress_bar: bool = True,
    ):
        """
        Initialize the FileManager with the given API URL, wallet path, and logger.

        Args:
            gateways (List[str] | str): A list of gateway URLs to use for file
            operations.
            wallet_path (str): The path to the wallet file.
            logger (Callable[[str], None]): A logging function.
            max_upload_size (int): The maximum size of a file to upload in bytes.
            show_progress_bar (bool): Whether to show a progress bar for file operations.
        """

        self.irys_uploader = init_irys(wallet=wallet, token=token,
                                       bundler_url=bundler_url)
        self.logger = logger
        self.max_upload_size = max_upload_size
        self.gateway_url = gateway_url
        self.show_progress_bar = show_progress_bar
        self.gateway_client = Node(self.gateway_url)
        self.bundler_client = Node(self.irys_uploader.url)
        if show_progress_bar:
            # Disable logging when using the progress bar
            def null(a: str, *b: Any) -> None:
                pass

            log.info = null  # type: ignore
            log.error = null  # type: ignore
            log.debug = null  # type: ignore

    @property
    def irys(self) -> Uploader:
        return self.irys_uploader

    @retry(exceptions=(Exception,), tries=5, delay=0.1)  # type: ignore
    def get_largefile_manifest(self, txid: str) -> Optional[LargeFileManifest]:
        """
        For handling of very large files, we chunk those files into different segments
        and upload those segments individually. We then aggregate the transaction IDs
        of each segment into a manifest file. This method retrieves the manifest file
        and returns a LargeFileManifest object containing the file data.

        Returns:
            LargeFileManifest: The LargeFileManifest object containing the file data.
        """
        # check if the file is ready
        self.check_ready(txid)

        # Check if the transaction is a manifest
        peer = self.bundler_client

        log.info(f"Getting Manifest file {peer.api_url} - {txid}")

        # # query_str = (
        # #     """
        # #             query {
        # #                 transaction(
        # #                     id: "%s"
        # #                 ) {
        # #                     id
        # #                     owner {
        # #                       address
        # #                     }
        # #                     tags {
        # #                       name
        # #                       value
        # #                     }
        # #                     block {
        # #                       timestamp
        # #                       height
        # #                     }
        # #                 }
        # #             }
        # #         """
        # #     % txid
        # # )
        # query_str = (
        #     """
        #             query {
        #                 transactions(
        #                     ids: ["%s"]
        #                 ) {
        #                     id
        #                     owner {
        #                       address
        #                     }
        #                     tags {
        #                       name
        #                       value
        #                     }
        #                     block {
        #                       timestamp
        #                       height
        #                     }
        #                 }
        #             }
        #         """
        #     % txid
        # )

        # try:
        #     res = peer.graphql(query_str)
        # except Exception as e:
        #     self.logger(f"Failed to download {txid}. Error: {e}")
        #     raise e

        # tx_tags: dict[str, str] = get_tags_dict(
        #     res["data"]["transaction"]["tags"])
        tx_tags = get_tags_dict(peer.tx_tags(txid))
        if tx_tags.get("Content-Type") != MANIFEST_CONTENT_TYPE:
            return None

        # construct the manifest dictionary from file bytes
        m_bytes = peer.tx_data(txid)
        _files_data = cast(Dict[str, Any], json.loads(m_bytes.decode()))
        _paths = _files_data.get("paths", {})

        tags = tx_tags

        # get the overall file size (useful for the progress bars)
        size_tags = [tag for tag in tags if tag["name"] == b"File-Size"]
        size_tag = (
            size_tags[0]["value"]
            if size_tags
            else len(_files_data.keys()) * self.max_upload_size
        )

        log.debug("manifest_data", _files_data)

        return LargeFileManifest(
            size=int(size_tag),
            files={k: _paths[k]["id"] for k in _paths.keys()},
        )

    def concat_files(self, temp_dir: Path, target_dir: Path) -> None:
        """
        Concatenate the sections into a single file.

        Args:
            temp_dir (Path): The temporary directory containing the sections.
            target_dir (Path): The target directory to write the concatenated file.

        Returns:
            None
        """
        # Concatenate the sections into a single file
        with open(target_dir, "wb") as file_handler:
            _iterdir = sorted(temp_dir.iterdir(), key=lambda x: int(x.stem))
            for section_path in _iterdir:
                with open(section_path, "rb") as section_handler:
                    file_handler.write(section_handler.read())

    def download(self, target_path: str | Path, txid: str) -> Path:
        """
        Download an Irys data transaction to a given path.

        Checks if the transaction is a manifest transaction with the type Sectioned.
        If so, downloads the manifest and then downloads each section. Otherwise,
        downloads the data transaction directly.

        Args:
            target_path (str|Path): The path to download the file to.
            txid (str): The transaction ID of the data transaction.

        Returns:
            Path: The absolute path of the downloaded file.
        """
        target_path = Path(target_path)

        log.info(f"Downloading {txid} to {target_path}")
        manifest_data = self.get_largefile_manifest(txid)
        progress = self.get_progress()

        if manifest_data is None:
            overall_task_id = progress.add_task(
                f"Downloading {target_path}",
                total=self.gateway_client.tx_data_size(txid),
            )

            return self.download_section(
                target_path,
                txid,
                self.gateway_url,
                progress,
                overall_task_id,
                overall_task_id,
            )

        overall_task_id = progress.add_task(
            f"Downloading {target_path}",
            total=manifest_data.size,
        )

        files: Dict[str, str] = manifest_data.files

        queue_processor: QueueProcessor[Gateway,
                                        Section] = QueueProcessor(logger=log)

        progress_bars: Dict[Gateway, PBarID] = {}

        for gateway in [self.gateway_url]:
            progress_bars[gateway] = progress.add_task(
                description=f"{STANDBY} - {gateway}",
                total=1,
                completed=0,
            )

        def worker_fn(gw: Gateway, task: Section) -> None:
            """
            Download a section of the file.

            Args:
                gw: The gateway to download from.
                task: The section to download.

            Returns:
                None
            """
            _section_path, _tx = task
            pbar_id = progress_bars[gw]
            progress.update(
                pbar_id,
                description=f"⬇️ - {gw}: {_section_path.name}",
                completed=0,
            )
            self.download_section(
                _section_path, _tx, gw, progress, pbar_id, overall_task_id
            )
            progress.update(
                pbar_id,
                description=f"{STANDBY} - {gateway}",
            )
            log.info(f"✅ Downloaded {_section_path} - {_tx}")

        def evict_hook(gw: Gateway) -> None:
            """
            Remove the progress bar for the given gateway. This is used after a worker
            has completed its download task.

            Args:
                gw: The gateway to remove the progress bar for.

            Returns:
                None
            """
            try:
                progress.remove_task(progress_bars[gw])
            except Exception:
                pass

        with tempfile.TemporaryDirectory() as _temp:
            temp_dir = Path(_temp)
            for _name, _tx_id in files.items():
                queue_processor.add_task((temp_dir / _name, _tx_id))

            for gateway in [self.gateway_url]:
                worker = Worker(context=gateway, work=worker_fn,
                                evict_hook=evict_hook)
                queue_processor.add_worker(worker)

            if self.show_progress_bar:
                with progress:
                    queue_processor.process()
            else:
                queue_processor.process()

            self.concat_files(temp_dir, Path(target_path))

        return Path(target_path)

    @retry(exceptions=(Exception,), tries=3, delay=0.1)  # type: ignore
    def check_ready(self, txid: str) -> str:
        """
        Check if the file is ready for download.

        Args:
            txid (str): The transaction ID of the data transaction.

        Returns:
            str: The gateway URL to download the file from.

        Raises:
            FileNotReadyException: If the file is pending and not ready for download.
        """
        gw = self.gateway_url

        log.debug(f"checking if resource is ready: {gw} {txid}")
        r = requests.get(f"{gw}/tx/{txid}")
        if r.status_code == 404:
            self.logger(f"file with txid {txid} does not exist")
            raise FileNotReadyException(
                f"File with txid {txid} does not exist on the gateway."
            )

        return gw

    def download_section(
        self,
        pathname: Path | str,
        txid: str,
        gateway: str,
        progress: Progress,
        pbar_id: PBarID,
        overall_task_id: PBarID,
    ) -> Path:
        """
        Download an Irys data transaction to a given path.

        Args:
            pathname (Path | str): The path to download to.
            txid (str): The transactionk ID of the data transaction.
            gateway (str): The gateway to download the file from.
            progress (Progress): A rich Progress object to display download progress.
            pbar_id (PBarID): The task ID of the current download progress.
            overall_task_id (PBarID): The task ID of the overall download progress.

        Returns:
            str: The absolute path of the downloaded file.

        Raises:
            FileNotReadyException: If the file is pending and not ready for download.
        """

        # Check if the file is pending
        self.check_ready(txid)

        loaded_bytes = 0
        gateway_client = self.gateway_client
        total_size = gateway_client.tx_data_size(txid)
        progress.update(pbar_id, total=total_size)
        with open(pathname, "wb") as binary_file:
            # Try downloading the transaction data directly from the default data
            # endpoint
            data = gateway_client.data(txid)

            # Write downloaded file to disk
            binary_file.write(data)

            progress.update(pbar_id, advance=len(data))
            progress.update(overall_task_id, advance=len(data))

            return Path(pathname)

    def get_progress(self) -> Progress:
        class SpeedColumn(ProgressColumn):
            """Custom column to display the upload speed."""

            def render(self, task: Task) -> str:
                speed = task.speed
                if speed is None:
                    return "N/A"
                speed_mbs = speed / (1024 * 1024)
                return f"{speed_mbs:.2f} MB/s"

        progress = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            "[progress.percentage]{task.percentage:>3.0f}%",
            SpeedColumn(),
            TimeElapsedColumn(),
            TimeRemainingColumn(),
        )
        return progress

    def upload(self, _file_path: Path | str, tags_dict: dict[str, str]):
        """
        Upload a file to Irys with the given tags. If the file is larger than
        the maximum node upload size, upload it in section.

        1. Split the file into multiple files each with a maximum size of
        MAXIMUM_UPLOAD_SIZE
        2. Upload each section to Irys using the upload_section method
        3. Upload a manifest containing the transaction IDs of each section
        4. Return the transaction ID of the manifest file

        Args:
            _file_path (Path | str): The path to the file to be uploaded.
            tags_dict (dict[str, str]): A dictionary of tags to be added to the
            transaction.
        """
        file_path: Path = (
            Path(_file_path) if isinstance(_file_path, str) else _file_path
        )

        progress = self.get_progress()

        overall_pbar_id = progress.add_task(
            f"Uploading {file_path.name}", total=os.path.getsize(file_path)
        )

        if os.path.getsize(file_path) < self.max_upload_size:
            return self.upload_section(
                file_path,
                tags_dict,
                progress,
                overall_pbar_id,
                overall_pbar_id,
            )

        # create a temporary directory to store the file sections
        log.info("Uploading large file in sections")
        with tempfile.TemporaryDirectory() as _temp:
            temp_dir = Path(_temp)
            # split the file into sections
            with open(file_path, "rb") as file_handler:
                file_number = 0
                while True:
                    chunk = file_handler.read(self.max_upload_size)
                    if not chunk:
                        break
                    section_dir = temp_dir / f"{file_number}.section"
                    with open(section_dir, "wb") as section_handler:
                        section_handler.write(chunk)
                    file_number += 1

            queue_processor: QueueProcessor[Gateway,
                                            Path] = QueueProcessor(logger=log)

            tx_ids: Dict[str, str] = {}

            log.info(f"Created {len(list(temp_dir.iterdir()))} sections")

            progress_bars: Dict[Gateway, PBarID] = {}

            for gateway in [self.gateway_url]:
                progress_bars[gateway] = progress.add_task(
                    f"{STANDBY} - {gateway}", total=1, completed=0
                )

            for section_path in temp_dir.iterdir():
                queue_processor.add_task(section_path)

            def worker_fn(gw: Gateway, _section_path: Path) -> None:
                pbar_id = progress_bars[gw]
                progress.update(
                    pbar_id,
                    description=f"⬆️ - {gw}: {_section_path.name}",
                    total=os.path.getsize(_section_path),
                    completed=0,
                )
                tx = self.upload_section(
                    _section_path, tags_dict, gw, progress, pbar_id, overall_pbar_id
                )
                tx_ids[_section_path.name] = tx.id
                log.info(f"✅ Uploaded {_section_path} - {tx.id}")
                progress.update(
                    pbar_id,
                    description=f"{STANDBY} - {gw}",
                )

            def evict_hook(gw: Gateway) -> None:
                progress.remove_task(progress_bars[gw])

            for _gateway in [self.gateway_url]:
                worker = Worker(context=_gateway, work=worker_fn,
                                evict_hook=evict_hook)
                queue_processor.add_worker(worker)

            if self.show_progress_bar:
                with progress:
                    queue_processor.process()
            else:
                queue_processor.process()

            timestamp = time.time()

            ritual_tags: Tags = {
                "App-Name": "Ritual",
                "App-Version": VERSION,
                "File-Type": "Sectioned",
                "File-Size": str(os.path.getsize(file_path)),
                "Unix-Time": str(timestamp),
                "File-Name": str(file_path.name),
            }

            # create a manifest object containing the transaction IDs of each section
            manifest_dict = {}

            for _dir, txid in tx_ids.items():
                manifest_dict[_dir] = txid

            m = Manifest(manifest_dict)

            # upload the manifest
            @retry(tries=len([self.gateway_url]), delay=0.2)  # type: ignore
            def _send_manifest() -> DataItem:
                data = self.create_sign_data_item(m.tobytes(), {
                    "Content-Type": MANIFEST_CONTENT_TYPE,
                    "Type": "manifest",
                    **ritual_tags,
                })
                self.irys_uploader.uploader.upload_tx(data)

                return data

            return _send_manifest()

    def create_sign_data_item(self, data: bytes, tags: Tags, target: str = None, anchor: str = None) -> DataItem:
        signer = self.irys_uploader.token_config.get_signer()
        item = create_data(data, signer, from_tags_dict(tags), target,
                           anchor if anchor else randbytes(16).hex())
        sign(item, signer)
        return item

    @retry(tries=3, delay=0.2)
    def upload_data(
        self, data: bytes, additional_tags: Optional[dict[str, str]] = None, **upload_opts
    ) -> DataItem:
        """
        Upload data to Irys with the given tags.

        Args:
            data: The data to be uploaded.
            additional_tags: Additional tags to be added to the transaction.

        Returns:
            DataItem: The created and signed transaction.
        """
        additional_tags = additional_tags or {}

        ritual_tags: Tags = {
            "App-Name": "Ritual",
            "App-Version": VERSION,
        }

        item = self.create_sign_data_item(data, {
            **ritual_tags,
            "Content-Type": MANIFEST_CONTENT_TYPE,
            "Type": "manifest",
            **additional_tags,
        })
        self.irys_uploader.uploader.upload_tx(item, **upload_opts)
        return item

    def upload_dict(
        self, d: Dict[str, Any], additional_tags: Optional[dict[str, str]] = None
    ) -> DataItem:
        """
        Upload a dictionary to Irys with the given tags

        Args:
            d: The dictionary to be uploaded.
            additional_tags: Additional tags to be added to the transaction.

        Returns:
            Transaction: The created and signed transaction.
        """
        return self.upload_data(json.dumps(d).encode(), additional_tags)

    def download_dict(self, txid: str) -> Dict[str, Any]:
        """
        Download a dictionary from Irys.

        Args:
            txid: The transaction ID of the data transaction.

        Returns:
            Dict[str, Any]: The downloaded dictionary.
        """
        return cast(Dict[str, Any], json.loads(self.download_data(txid)))

    def upload_manifest(
        self,
        manifest_dict: dict[str, str],
        additional_tags: Optional[dict[str, str]] = None,
    ) -> DataItem:
        """
        Upload a manifest to Irys with the given tags

        Args:
            manifest_dict: The manifest dictionary to be uploaded.
            additional_tags: Additional tags to be added to the transaction.

        Returns:
            Transaction: The created and signed transaction.
        """
        return self.upload_data(Manifest(manifest_dict).tobytes(), additional_tags)

    def download_data(self, txid: TxID) -> bytes:
        """
        Download data from Irys.

        Args:
            txid: The transaction ID of the data transaction.

        Returns:
            bytes: The downloaded data.
        """
        return cast(bytes, self.gateway_client.tx_data(ensure_str(txid)))

    def upload_section(
        self,
        file_path: Path,
        tags_dict: dict[str, str],
        progress: Progress,
        task_id: PBarID,
        overall_task_id: PBarID,
    ):
        """
        Upload a file to Irys with the given tags.

        Args:
            file_path (Path): The path to the file to be uploaded.
            tags_dict (dict[str, str]): A dictionary of tags to be added to the
            transaction.
            gateway (str): The gateway to upload the file to.
            progress (Progress): A rich Progress object to display upload progress.
            task_id (TaskId): The task ID of the current upload progress.
            overall_task_id (TaskId): The task ID of the overall upload progress.

        Returns:
            DataItem: The created and signed transaction.
        """

        default_tags: Dict[str, str] = {
            "File-Size": str(os.path.getsize(file_path)),
        }

        total_tags = {**tags_dict, **default_tags}

        with open(file_path, "rb", buffering=0) as file_handler:
            data = file_handler.read()
            tx = self.create_sign_data_item(data, total_tags)
            self.irys_uploader.uploader.upload_tx(tx)
            progress.update(overall_task_id, advance=int(len(data)))

        return tx

    def file_exists(self, file_path: str, txid: str) -> bool:
        """
        Given a local file path and a transaction ID, check if the file exists on
        Irys. Checks for the following:
        - Local file exists
        - Local file's size matches transaction data size
        - Local file's sha256 digest matches transaction digest

        Args:
            file_path (str): The path to the local file.
            txid (str): The transaction ID to check against.

        Returns:
            bool: True if the file exists and matches the transaction, False otherwise.
        """
        # query_str = (
        #     """
        #         query {
        #             transaction(
        #             id: "%s"
        #             )
        #             {
        #                 owner{
        #                     address
        #                 }
        #                 data{
        #                     size
        #                     type
        #                 }
        #                 tags{
        #                     name
        #                     value
        #                 }
        #             }
        #         }
        #     """
        #     % txid
        # )

        # res = self.bundler_client.graphql(query_str)

        # tx_file_size: int = int(res["data"]["transaction"]["data"]["size"])
        # tx_tags: dict[str, str] = get_tags_dict(
        #     res["data"]["transaction"]["tags"])
        tx_file_size = self.gateway_client.tx_field(txid, "data_size")
        tx_tags = self.gateway_client.tx_field(txid, "tags")

        local_file_exists, size_matches, digest_matches = (
            False,
            False,
            False,
        )

        def _log() -> None:
            """Log the current status of file existence checks."""
            self.logger(
                f"file_path={file_path} local_file_exists={local_file_exists} "
                f"size_matches={size_matches} digest_matches={digest_matches}",
            )

        local_file_exists = os.path.exists(file_path)

        if not local_file_exists:
            _log()
            return False

        size_matches = tx_file_size == os.path.getsize(file_path)

        if not size_matches:
            _log()
            return False

        digest_matches = tx_tags.get(
            "File-SHA256") == get_sha256_digest(file_path)
        _log()
        return digest_matches



import logging

from ritual_irys.cli import cli


def main() -> int:
    from nicelog import setup_logging  # type: ignore

    setup_logging()

    logging.getLogger("urllib3").setLevel(logging.CRITICAL)

    cli()
    return 0



"""
Utility functions for uploading/downloading repositories to/from Irys. Each repository
contains multiple artifact files.


Repository files (artifacts) are logically grouped together via a Manifest file, which
maps individual transaction data to named files.

When uploading a repo directory, a version mapping dictionary file is expected to be
provided. The mapping should contain a map of filename to version tag. The version tag
is useful if a specific version of a file is meant to be downloaded. If no mapping is
specified, version will be an empty string.
"""

import json
import logging
import mimetypes
import os
import time
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, cast

import requests
from pydantic import BaseModel
from requests.exceptions import HTTPError
from ritual_irys.file_manager import FileManager
from ritual_irys.gateway import Node
from ritual_irys.manifest import MANIFEST_CONTENT_TYPE, Manifest
from ritual_irys.types import IrysRepoId, Tags
from ritual_irys.utils import edge_unix_ts, from_tags_dict, get_sha256_digest
from ritual_irys.version import VERSION

log = logging.getLogger(__name__)


class UploadIrysRepoResult(BaseModel):
    """Model to represent the result of a repository upload."""

    repo_id: IrysRepoId
    transaction_id: str
    manifest_url: str


class RepoManager(FileManager):
    def download_artifact_file(
        self,
        repo_id: Union[IrysRepoId, str],
        file_name: str,
        version: Optional[str] = None,
        force_download: bool = False,
        base_path: str | Path = ".",
    ) -> Path:
        """Downloads a specific artifact from Irys.

        Args:
            repo_id (Union[IrysRepoId, str]): id of the repo, if provided as a string, the
                format must be of the form `owner`/`name`. Where `owner` is the wallet
                address of the uploader and `name` is the repository's name.
            file_name (str): name of artifact
            version (Optional[str], optional): Version of file. Defaults to None.
                If none specified, will fetch the latest version.
            force_download (bool, optional): If True, will download file even if it
                already exists at that path. Defaults to False.
            base_path (str, optional): path to download file to. Defaults to "".

        Raises:
            ValueError: if wallet file path is not specified or wallet file is not found.

        Returns:
            Path: path to downloaded file
        """
        if isinstance(repo_id, str):
            repo_id = IrysRepoId.from_str(repo_id)

        base = Path(base_path)
        os.makedirs(base, exist_ok=True)
        owners = [repo_id.owner]

        file_version_str = (
            ""
            if not version
            else """
            {
                name: "File-Version",
                values: ["%s"]
            },
        """
            % version
        )
        query_str = """
        query {
            transactions(
                order:DESC,
                owners: %s,
                tags: [
                    {
                        name: "App-Name",
                        values: ["Ritual"]
                    },
                    %s
                    {
                        name: "File-Name",
                        values: ["%s"]
                    },
                    {
                        name: "Repo-Name",
                        values: ["%s"]
                    }
                ])
            {
                edges {
                    node {
                        timestamp
                        id
                        address
                        tags {
                            name
                            value
                        }
                    }
                }
            }
        }
        """ % (
            json.dumps(owners),
            file_version_str,
            file_name,
            repo_id.name,
        )
        log.debug(query_str)

        file_path: Path = base.joinpath(file_name)

        res = self.bundler_client.graphql(query_str)
        transactions = res["data"]["transactions"]["edges"]
        transactions.sort(reverse=True, key=edge_unix_ts)

        if len(transactions) == 0:
            raise ValueError(
                f"Could not find any matching artifacts for: "
                f"({repo_id}, {file_name}, {version or 'latest'})"
            )

        transaction = transactions[0]

        tx_metadata: dict[str, Any] = transaction["node"]

        tx_id = tx_metadata["id"]

        if force_download or not self.file_exists(str(file_path), tx_id):
            log.info(f"downloading {tx_metadata}")
            return self.download(str(file_path), tx_id)
        else:
            log.info(f"not downloading {
                     tx_metadata} because it already exists")
            return Path(file_path)

    def upload_repo(
        self,
        name: str,
        path: Path | str,
        version_mapping_file: Optional[str] = None,
        version_mapping: Optional[Dict[str, str]] = None,
        extra_file_tags: Optional[Dict[str, Tags]] = None,
    ) -> UploadIrysRepoResult:
        """
        Uploads a repo directory to Irys. For every repository upload, a manifest
        mapping is created.

        Args:
            name (str): Name of the repository. Once uploaded, the repo will be
                accessible via the repo Id: `owner/name`. Where `owner` is the wallet
                address of the uploader and `name` is the repository's name.
            path (Path | str): Path to the directory containing the repository files.
            version_mapping_file (str): Path to a json dict file mapping file names to
                specific versions. If a specific mapping is found, the File-Version
                attribute is tagged with the value. This is to facilitate uploading and
                downloading version specific files.
            version_mapping (dict[str, str]): Dictionary mapping file names to specific
                versions. If a specific mapping is found, the File-Version attribute is
                tagged with the value. This is to facilitate uploading and downloading
                version specific files. If provided, this will override the version_file.
            extra_file_tags (dict[str, Tags]): Dictionary mapping file names to
                additional tags to be added to the file. This is useful for adding
                additional metadata to each file.

        Raises:
            ValueError: if wallet file path is not specified or wallet file is not found.

        Returns:
            UploadIrysRepoResult: Result of the upload containing repo_id,
            transaction_id, and manifest_url.
        """

        # path to load files from
        _path: Path = Path(path)

        # load all sub-paths in this path
        p = _path.glob("**/*")

        # get timestamp to tag files with
        timestamp = time.time()

        # filter out simlinks and non-files
        files = [x for x in p if x.is_file()]

        _version_mapping = {}
        if version_mapping:
            _version_mapping = version_mapping
        elif version_mapping_file:
            with open(version_mapping_file, "r") as vf:
                _version_mapping = json.load(vf)
        self.logger(f"using mapping {_version_mapping}")

        # keep track of entries via a manifest
        manifest_dict: dict[str, str] = {}

        ritual_tags: Tags = {
            "App-Name": "Ritual",
            "App-Version": VERSION,
            "Unix-Time": str(timestamp),
            "Repo-Name": str(name),
        }

        for f in files:
            rel_path = os.path.relpath(f, _path)

            self.logger(f"looking at {f} ({rel_path}) Size: {
                        os.path.getsize(f)}")

            content_type = (
                guess
                if (guess := mimetypes.guess_type(f)[0])
                else "application/octet-stream"
            )
            file_extra_tags = (
                extra_file_tags.get(rel_path, {}) if extra_file_tags else {}
            )

            tags_dict: Tags = {
                **file_extra_tags,
                "Content-Type": content_type,
                "File-Version": _version_mapping.get(str(rel_path), "0.0.0"),
                "File-Name": rel_path,
                "File-SHA256": get_sha256_digest(str(f)),
                **ritual_tags,
            }

            self.logger(f"uploading: {f} with tags: {tags_dict}")

            tx = self.upload(f, tags_dict)

            # we are done uploading the whole file, keep track if filename -> tx.id
            manifest_dict[str(os.path.relpath(f, _path))] = tx.id
            self.logger(f"uploaded file {f} with id {
                        tx.id} and tags {tags_dict}")

        # we create a manifest of all the files to their transactions
        m = Manifest(manifest_dict)
        t = self.create_sign_data_item(m.tobytes(), tags={
            "Content-Type": MANIFEST_CONTENT_TYPE,
            "Type": "manifest",
            **ritual_tags,
        })
        self.irys_uploader.uploader.upload_tx(t)
        # upload the manifest
        # t = Transaction(self.irys, peer=self.peer, data=m.tobytes())

        # t.add_tags(
        #     {
        #         "Content-Type": MANIFEST_CONTENT_TYPE,
        #         "Type": "manifest",
        #         **ritual_tags,
        #     }
        # )

        # t.sign()
        # t.send()
        # t = self.upload(m.tobytes(), from_tags_dict({
        #     "Content-Type": MANIFEST_CONTENT_TYPE,
        #     "Type": "manifest",
        #     **ritual_tags,
        # }))

        self.logger(f"uploaded manifest with tx id {t.id}")

        return UploadIrysRepoResult(
            repo_id=IrysRepoId(owner=self.irys_uploader.address, name=name),
            transaction_id=t.id,
            manifest_url=f"{self.gateway_url}/{t.id}",
        )

    def download_file_in_repo(
        self,
        repo_id: Union[IrysRepoId, str],
        file_name: str,
        base_path: str | Path = Path("."),
    ) -> Path:
        """
        Downloads a file from a repo on Irys.

        Args:
            repo_id: Irys repo id, in the format of `owner`/`name` where `owner` is
                the wallet address of the uploader and `name` is the repository's name.
            file_name: name of the file to download.
            base_path: path to download the file to.

        Returns:
            path to the downloaded file.

        """

        base_path = Path(base_path)
        manifest = self.get_repo_manifest(repo_id)
        file_tid = manifest["paths"][file_name]["id"]
        return self.download(base_path / file_name, file_tid)

    def get_repo_manifest(self, repo_id: Union[IrysRepoId, str]) -> dict[str, Any]:
        """
        Get the manifest of a repo from Irys.
        Args:
            repo_id: Irys repo id, in the format of `owner`/`name` where `owner` is
                the wallet address of the uploader and `name` is the repository's name.

        Returns:
            dict: manifest of the repo.

        """

        if isinstance(repo_id, str):
            repo_id = IrysRepoId.from_str(repo_id)
        owners = [repo_id.owner]
        query_str = """
        query {
            transactions(
                order:DESC,
                owners: %s,
                tags: [
                    {
                        name: "App-Name",
                        values: ["Ritual"]
                    },
                    {
                        name: "Repo-Name",
                        values: ["%s"]
                    },
                    {
                        name: "Type",
                        values: ["manifest"]
                    }
                ]
            )
            {
                edges {
                    node {
                        timestamp
                        id
                        address
                        tags {
                            name
                            value
                        }
                    }
                }
            }
        }
        """ % (
            json.dumps(owners),
            repo_id.name,
        )

        # self.logger(query_str)
        log.info("getting first query")
        try:
            res = self.bundler_client.graphql(query_str)
        except HTTPError as e:
            raise ValueError(f"Error querying repo manifests: {e}")
        log.info("done getting first query")

        # get latest Manifest

        manifests = res["data"]["transactions"]["edges"]
        manifests.sort(reverse=True, key=edge_unix_ts)
        self.logger(f"found {len(manifests)} manifests for {repo_id}")

        if len(manifests) == 0:
            raise ValueError(
                "Could not find any matching repo manifests from query.")

        manifest = manifests[0]

        tx_id = manifest["node"]["id"]

        # download manifest data
        self.logger(f"found manifest {manifest}")

        log.info("getting manifest")
        try:
            m = json.loads(self.gateway_client.tx_data(tx_id))
        except Exception as e:
            log.info("Exception while getting manifest")
            r = requests.get(f"{self.gateway_client.api_url}/tx/{tx_id}")
            raise ValueError(f"Error fetching manifest data: {e}")
        log.info("done getting manifest")

        self.logger(f"loaded manifest {m}")
        return cast(dict[str, Any], m)

    def download_repo(
        self,
        repo_id: Union[IrysRepoId, str],
        base_path: str | Path = Path("."),
        force_download: bool = False,
    ) -> List[Path]:
        """Downloads a repo from Irys to a given directory.

        Args:
            repo_id (Union[IrysRepoId, str]): id of the repo, if provided as a string, the
                format must be of the form `owner`/`name`. Where `owner` is the wallet
                address of the uploader and `name` is the respository's name.
            base_path (str | Path, optional): path to download files to. Defaults to ".".
            force_download (bool, optional): If True, will download files even if they
                already exist. Defaults to False.

        Raises:
            ValueError: if wallet file path is not specified or wallet file is not found.
            ValueError: if matching repo manifest not found

        Returns:
            list[str]: downloaded file paths
        """

        manifest = self.get_repo_manifest(repo_id)

        base = Path(base_path)
        base.mkdir(parents=True, exist_ok=True)

        paths = []

        with ThreadPoolExecutor() as executor:
            for pathname, tid in manifest["paths"].items():
                file_tid: str = tid["id"]
                joined_path: Path = base.joinpath(pathname)

                # check if file exists
                if force_download or not self.file_exists(str(joined_path), file_tid):
                    st = time.time()
                    self.logger(f"downloading file {pathname} for {file_tid}")
                    paths.append(
                        executor.submit(self.download, str(
                            joined_path), file_tid)
                    )
                    self.logger(f"downloaded in {
                                time.time() - st} sec: {joined_path}")
                else:
                    self.logger(
                        f"Path {
                            joined_path} already exists and will not be downloaded. "
                        + "Please remove it or use --force_download flag."
                    )

        return [p.result() for p in paths]
